%A demo of conservation laws modeling traffic flow
%Uses Lax-Friedrichs scheme
%Author: Jeff Calder, 2016

N = 1000; %Number of grid points
k = 500; %number of steps between frames
T = 100000; %Number of iterations
h = 1/(N-1); %Grid spacing
dt = h/10; %Time step for solving pde

x = 0:1/(N-1):1;
cd = 75; %car density
cs = 150; %size of cars
y_max = 1; %Plot Axes
y_min = -0.1;
ps = 0; %pause duration (to slow down animation)
var = 0.5; %variance in position of cars (var = 0 means no randomness, max=1)


%Initial condition
u0 = 0.3*ones(1,length(x));
u = u0;

%Initialize cars according to initial density u0
a = 0;
b = (exprnd(1)-1)*var+1;
car_pos = [];
for k=1:length(x)
   a = a + h*cd*u0(k);
   if(a >= b)
      a=0;
      b = (exprnd(1)-1)*var+1;
      car_pos = [car_pos, k*h]; % Add new car
   end
end

if(isempty(car_pos)) %If cars happen to be empty, add one at x=h
   car_pos = h;
end

%Draw initial plot
plot(x,u,'linewidth',2);
ylim([y_min y_max]);
title('Traffic Flow');
xlabel('Position');
ylabel('Car density');
set(gca,'fontsize',15);
hold;
scatter(car_pos,zeros(1,length(car_pos)),cs,'square');
hold;
drawnow;
pause;

%Main loop for solving PDE
d=1;
i=0;
b = (exprnd(1)-1)*var+1;
while(d>0)
   i = i+1;

   %Draw plot every k iterations
   if mod(i,k) == 0
      plot(x,u,'linewidth',2);
      ylim([y_min y_max]);
      title('Traffic Flow');
      xlabel('Position');
      ylabel('Car density');
      set(gca,'fontsize',15);
      hold;
      scatter(car_pos,zeros(1,length(car_pos)),cs,'square');
      hold;
      drawnow;
      pause(ps);
   end
   
   %Update position of cars
   car_vel = 1-u(ceil(car_pos/h));
   car_pos = car_pos + dt*car_vel;
   car_pos(car_pos > 1-h) = []; %remove cars exiting on right

   %Add cars randomly on left
   a = sum(u(1:ceil(min(car_pos)/h)))*cd*h;
   if(a >= b)
      car_pos = [h,car_pos]; %add car at x=h
      a = 0;
      b = (exprnd(1)-1)*var+1;
   end

   %Update PDE
   un = [u([2:end]),(1-cos(i*dt))/2]; %Boundary conditions are encoded here
   up = [0.3,u([1:end-1])];

   %Update solution of PDE using Lax-Friedrichs scheme
   u_new = u - dt*(u - up + (up.^2-un.^2)/2)/h;
   d = max(abs(u-u_new));
   u = u_new;
end

%Draw final plot if simulation converged to steady state
plot(x,u,'color','red','linewidth',2);
ylim([y_min y_max]);
title('Traffic flow');
xlabel('Position');
ylabel('Car density');
set(gca,'fontsize',15);
hold;
scatter(car_pos,zeros(1,length(car_pos)),cs,'square');
hold;
pause;


